package com.kalixia.intellij.plugins.je.browser;

import com.intellij.openapi.Disposable;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.project.Project;
import com.intellij.psi.PsiClass;
import com.intellij.psi.PsiManager;
import com.intellij.psi.search.GlobalSearchScope;
import com.kalixia.intellij.plugins.je.browser.actions.AddStoreAction;
import com.kalixia.intellij.plugins.je.browser.actions.CloseStoreAction;
import com.kalixia.intellij.plugins.je.browser.actions.RemoveStoreAction;
import com.kalixia.intellij.plugins.je.browser.actions.OpenStoreAction;
import com.kalixia.intellij.plugins.je.browser.table.JEEntityTableModel;
import com.kalixia.intellij.plugins.je.browser.tree.EntityTreeNode;
import com.kalixia.intellij.plugins.je.browser.tree.JEBrowserTreeCellRenderer;
import com.kalixia.intellij.plugins.je.browser.tree.PrimaryKeyMetadataTreeNode;
import com.kalixia.intellij.plugins.je.browser.tree.SecondaryKeyMetadataTreeNode;
import com.kalixia.intellij.plugins.je.browser.tree.StoreTreeNode;
import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.Environment;
import com.sleepycat.je.EnvironmentConfig;
import com.sleepycat.persist.StoreConfig;
import com.sleepycat.persist.model.EntityMetadata;
import com.sleepycat.persist.model.EntityModel;
import com.sleepycat.persist.model.PrimaryKeyMetadata;
import com.sleepycat.persist.model.SecondaryKeyMetadata;
import com.sleepycat.persist.raw.RawStore;
import org.jdesktop.swingx.JXErrorPane;
import org.jdesktop.swingx.JXTable;
import org.jdesktop.swingx.JXTree;
import org.jdesktop.swingx.decorator.HighlighterFactory;
import org.jdesktop.swingx.error.ErrorInfo;
import org.jdesktop.swingx.error.ErrorLevel;
import org.jetbrains.annotations.NotNull;
import javax.swing.*;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableModel;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.File;
import static java.lang.String.format;
import java.util.Enumeration;
import java.util.Map;
import java.util.Set;

public class JEBrowserPanel implements Disposable {
    private JTabbedPane tabbedPane1;
    private JPanel rootComponent;
    private JTree storesTree;
    private DefaultTreeModel storesTreeModel;
    private JTable entitiesTable;
    private JPopupMenu storesMenu, storeMenu;
    private final Project project;
    private final Settings settings;
    private Logger logger = Logger.getInstance(getClass().getName());

    public JEBrowserPanel(Project project, final Settings settings) {
        this.project = project;
        this.settings = settings;
    }

    public JPanel getRootComponent() {
        return rootComponent;
    }

    @SuppressWarnings({"AsymmetricFieldRead"})
    private void createUIComponents() {
        DefaultMutableTreeNode rootNode = new DefaultMutableTreeNode("Stores");

        storesTreeModel = new DefaultTreeModel(rootNode);
        storesTree = new JXTree(storesTreeModel);
        storesTree.setCellRenderer(new JEBrowserTreeCellRenderer());
        storesTree.addMouseListener(new MyTreeMouseListener());
        storesTree.addTreeSelectionListener(new TreeSelectionListener() {
            public void valueChanged(TreeSelectionEvent event) {
                TreePath path = event.getPath();
                TreeNode node = (TreeNode) path.getLastPathComponent();
                if (node instanceof EntityTreeNode) {
                    EntityTreeNode entityNode = ((EntityTreeNode) node);
                    String clazz = entityNode.getClazz();
                    final TableModel model = new JEEntityTableModel(project, entityNode.getStore(), clazz);
//                    final TableModel model = new JEEntityLargeTableModel(project, entityNode.getStore(), clazz);
                    entitiesTable.setModel(model);
                }
            }
        });

        entitiesTable = new JXTable();
        entitiesTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        ((JXTable) entitiesTable).setColumnControlVisible(true);
        ((JXTable) entitiesTable).setSortable(false);
        ((JXTable) entitiesTable).setHighlighters(HighlighterFactory.createAlternateStriping());

        buildPopupMenus();
    }

    private void populateStore(StoreTreeNode storeNode) throws ClassNotFoundException {
        RawStore store = storeNode.getStore();
        EntityModel model = store.getModel();
        Set<String> classes = model.getKnownClasses();
        GlobalSearchScope projectScope = project.getProjectScope();
        for (String clazz : classes) {
            if (logger.isDebugEnabled()) logger.debug(format("Processing class %s\r", clazz));
            PsiClass psiClass = PsiManager.getInstance(project).findClass(clazz, projectScope);
            // skip non-project classes
            if (psiClass == null)
                continue;
            // skip non-entity classes
            if (!model.getClassMetadata(clazz).isEntityClass())
                continue;
            EntityTreeNode entityClassNode = new EntityTreeNode(store, clazz);
            storeNode.add(entityClassNode);

            EntityMetadata entityMetadata = model.getEntityMetadata(clazz);
            if (entityMetadata == null)
                continue;
            PrimaryKeyMetadata primaryKey = entityMetadata.getPrimaryKey();
            if (logger.isDebugEnabled()) logger.debug(format("Found primary key: %s\r", primaryKey.getName()));
            entityClassNode.add(new PrimaryKeyMetadataTreeNode(store, primaryKey));

            Map<String, SecondaryKeyMetadata> secondaryKeys = entityMetadata.getSecondaryKeys();
            for (SecondaryKeyMetadata secondaryKey : secondaryKeys.values()) {
                if (logger.isDebugEnabled())
                    logger.debug(format("Found secondary key: %s %s %s\r",
                            secondaryKey.getKeyName(),
                            secondaryKey.getRelationship(),
                            secondaryKey.getRelatedEntity()
                    ));
                entityClassNode.add(new SecondaryKeyMetadataTreeNode(store, secondaryKey));
            }
        }
    }

    private void buildPopupMenus() {
        storesMenu = new JPopupMenu();
        storesMenu.add(new AddStoreAction(this));

        storeMenu = new JPopupMenu();
        storeMenu.add(new OpenStoreAction(this));
        storeMenu.add(new CloseStoreAction(this));
        storeMenu.add(new RemoveStoreAction(this));
    }

    public void addStore(String directory, String name) throws DatabaseException, ClassNotFoundException {
        addStore(directory, name, true);
    }

    public void addStore(String directory, String storeName, boolean saveInSettings) throws DatabaseException, ClassNotFoundException {
        if (directory == null) {
            logger.error("You should select an environment directory!");
            return;
        }
        StoreTreeNode storeNode = openStore(directory, storeName);
        if (saveInSettings) {
            settings.addStore(directory, storeName);
        }
        DefaultMutableTreeNode rootNode = (DefaultMutableTreeNode) storesTreeModel.getRoot();
        rootNode.add(storeNode);
        storesTreeModel.nodeStructureChanged(rootNode);
    }

    public void openStore(@NotNull StoreTreeNode storeNode) throws DatabaseException, ClassNotFoundException {
        RawStore store = storeNode.getStore();
        openStore(storeNode, store.getEnvironment().getHome().getAbsolutePath(), store.getStoreName());
    }

    public StoreTreeNode openStore(@NotNull String directory, String storeName) throws DatabaseException, ClassNotFoundException {
        StoreTreeNode storeNode = new StoreTreeNode();
        openStore(storeNode, directory, storeName);
        return storeNode;
    }

    private void openStore(@NotNull StoreTreeNode storeNode, @NotNull String directory, @NotNull String storeName) throws DatabaseException, ClassNotFoundException {
        EnvironmentConfig environmentConfig = new EnvironmentConfig();
        environmentConfig.setReadOnly(true);
        Environment environment = new Environment(new File(directory), environmentConfig);
        StoreConfig storeConfig = new StoreConfig();
        storeConfig.setReadOnly(true);
        RawStore store = new RawStore(environment, storeName, storeConfig);
        storeNode.setStore(store);
        storeNode.setOpened(Boolean.TRUE);
        try {
            populateStore(storeNode);
        } catch (ClassNotFoundException e) {
            JXErrorPane.showDialog(getRootComponent(), new ErrorInfo(
                    "Open Berkeley DB (Java Edition) store",
                    e.getMessage(),
                    null, "ERROR", e, ErrorLevel.SEVERE, null)
            );
            throw e;
        }
        if (storeNode.isLeaf())
            storesTree.scrollPathToVisible(new TreePath(storeNode.getPath()));
        else
            storesTree.scrollPathToVisible(new TreePath(((DefaultMutableTreeNode) storeNode.getChildAt(0)).getPath()));
    }

    public void closeStore(@NotNull TreePath storePath) {
        StoreTreeNode node = (StoreTreeNode) storePath.getLastPathComponent();
        RawStore store = node.getStore();
        try {
            entitiesTable.setModel(new DefaultTableModel());
            node.removeAllChildren();
            node.setOpened(Boolean.FALSE);
            storesTree.collapsePath(storePath);
            if (node.isOpened()) {
                Environment environment = store.getEnvironment();
                store.close();
                environment.close();
            }
        } catch (Exception e) {
            JXErrorPane.showDialog(e);
        }
    }

    public void removeStore(@NotNull TreePath storePath) {
        if (!(storePath.getLastPathComponent() instanceof StoreTreeNode))
            return;
        StoreTreeNode node = (StoreTreeNode) storePath.getLastPathComponent();
        RawStore store = node.getStore();
        try {
            entitiesTable.setModel(new DefaultTableModel());
            storesTreeModel.removeNodeFromParent(node);
            if (node.isOpened()) {
                Environment environment = store.getEnvironment();
                settings.removeStore(store.getStoreName());
                store.close();
                environment.close();
            }
            node.setOpened(Boolean.FALSE);
        } catch (Exception e) {
            JXErrorPane.showDialog(e);
        }
    }

    public JTree getStoresTree() {
        return storesTree;
    }

    public Project getProject() {
        return project;
    }

    public JPopupMenu getStoreMenu() {
        return storeMenu;
    }

    @SuppressWarnings("unchecked")
    public void dispose() {
        DefaultMutableTreeNode rootNode = (DefaultMutableTreeNode) storesTree.getModel().getRoot();
        Enumeration<StoreTreeNode> stores = rootNode.children();
        while (stores.hasMoreElements()) {
            StoreTreeNode storeNode = stores.nextElement();
            try {
                storeNode.getStore().close();
            } catch (DatabaseException e) {
                JXErrorPane.showDialog(e);
            }
        }
    }

    class MyTreeMouseListener extends MouseAdapter implements MouseListener {
        public void mousePressed(MouseEvent e) {
            maybeShowPopup(e);
        }

        public void mouseReleased(MouseEvent e) {
            maybeShowPopup(e);
        }

        private void maybeShowPopup(MouseEvent e) {
            if (e.isPopupTrigger()) {
                JXTree tree = (JXTree) e.getComponent();
                TreePath path = tree.getPathForLocation(e.getX(), e.getY());
                if (path == null) {
                    storesMenu.show(tree, e.getX(), e.getY());
                } else {
                    DefaultMutableTreeNode node = (DefaultMutableTreeNode) path.getLastPathComponent();
                    storesTree.setSelectionPath(path);
                    Object userObject = node.getUserObject();
                    if (userObject == "Stores")
                        storesMenu.show(tree, e.getX(), e.getY());
                    else if (node instanceof StoreTreeNode)
                        storeMenu.show(tree, e.getX(), e.getY());
                }
            }
        }
    }

}
